#include "stroj_casu.h"

void state_init(void)
{
//	status = (P_STATUS) xmalloc(sizeof(T_STATUS));
	_parse_barbar_init();
	_parse_command_line_init();
	_fortune_init();
	languages = 0;
	LANG_ON(BARBAR);
	mood = MOOD_START;
	updatemood();
	init_blinking();
	stroj_offset = 0;
}

int please(void)
{
	int retval;
	if ((mood > MOOD_MAX - MOOD_PLEASE_ADD) || (please_disgust > PLEASE_DISGUST_LIMIT))
	{
		// disgusted by a too polite user
		// keeping on pleasing isn't the right idea
		mood -= MOOD_DISGUSTED;
		if (mood < 0) mood = 0;
		retval = 0;
	}
	else
	{
		// Ok, I like being pleased
		mood += MOOD_PLEASE_ADD;
		retval = 1;
	}
	please_disgust++;
	updatemood();
	return retval;
}

int noplease(void)
{
	please_disgust = 0;
	if (mood <= MOOD_NOPLEASE_SUB)
	{
		// Too many rude users -- I won't do anything before they
		// please
		mood = 0;
		updatemood();
		return 0;
	}
	mood -= MOOD_NOPLEASE_SUB;
	please_disgust = 0;
	updatemood();
	return 1;
}

#define	ASORT_PREFIX(x)	_input_command_##x
#define ASORT_KEY_TYPE	char *
#define ASORT_LT(x,y)	(strcmp((x),(y))>0)
#include <ucw/sorter/array-simple.h>

void input_command(char * command_in)
{
	if (MAX_COMMANDS_PENDING <= commands_in_list)
		return;
	char * command;
	// Vytvorime trvale uloziste pro prikaz
	command = xmalloc_zero((strlen(command_in) + 2)*sizeof(char));
	strcpy(command, command_in);
	// Vlozime do pole a zatridime
	commands_todo[commands_in_list] = command;
	commands_in_list++;
	_input_command_sort(commands_todo, commands_in_list);
}

void show_todo_list(UNUSED char * trash)
{
	int i;
	for (i=0;i<commands_in_list;i++)
		printf("%d:\t%s\n", i, commands_todo[i]);
}

void loc_exec_command(UNUSED char * trash)
{
	char * command;
	// Vyjmeme prikaz z pole
	if (commands_in_list == 0) return; // neni tam nic? tak nic neudelame
	commands_in_list--;
	command = commands_todo[commands_in_list];
	printf("executing %s\n", command);
	// Pro kazdy jazyk zjistime, jestli ho zname, a pokud ano, pak v nem
	// spustime prikaz
	if (LANG_STAT(CZECH)) parse_czech(command);
	if (LANG_STAT(LATIN)) parse_latin(command);
	if (LANG_STAT(BARBAR)) parse_barbar(command);
	if (LANG_STAT(OLDIRISH)) parse_oldirish(command);
	if (LANG_STAT(OLDCZECH)) parse_oldczech(command);
	if (LANG_STAT(ENGLISH)) parse_english(command);
	// Prikaz uz dale nepotrebujeme, tak se ho zbavime
	xfree(command);
}


void play_sound(char * path)
{
	char command[SHORT_STRING];
	if (fork()) return;
	file_close_all();
	sprintf(command, "%s %s %s < /dev/null > /dev/null 2> /dev/null\n",
	    MEDIA_PLAYER, "-quiet", path);
	system(command);
	msg(L_DEBUG, command);
	exit(0);
}

void stop_sound(void)
{
	printf("Killing mplayer \n");
	system("killall mplayer");
}

struct {
	s64 dest;
	s64 timestamp;
} next_move;

struct main_timer *_planned_tasks_timer;

void set_next_move(s64 dest) { // standard UNIX timestamp
	next_move.dest = dest;
	next_move.timestamp = stroj_now;
	printf("set_next_move(%Ld)\n", dest);
}

s64 get_next_move(void) {
	if (next_move.timestamp + MOVEDEST_TIMEOUT < stroj_now)
		return stroj_now;
	return next_move.dest;
//	printf("get_next_move\n");
}

timestamp_t _oldest_time;

void _moved(struct main_timer *mt)
{
#define i ((int) mt->data)
	timer_add(mt, main_now + 500);
	set_orange_led(i & 1);
	set_green_led(i & 1);
	if (i < 4)
		set_big_red_led(1);
	else if (i < 10)
	{
		if (i == 4)
		{
			play_sound("ogg/time_transport.ogg");
			lcd_printf("TRANSPORT\n");
			lcd_printf("IN PROGRESS\n");
		}
		set_big_red_led(0);
		set_big_yellow_led(1);
	}
	else if (i < 30)
	{
		if (i == 10) play_sound("ogg/time_transport.ogg");
		set_big_yellow_led(0);
		set_big_green_led(1);
	}
	else
	{
		play_sound("ogg/time_transport.ogg");
		lcd_printf("TRANSPORT DONE\n");
		lcd_printf("TRANSPORT DONE\n");
		set_lamp(0);

		timer_del(mt);
		return;
	}

	mt->data++;
#undef i
}

int bagr;

void move_in_time(void) {
	if (bagr == 1) return;
	bagr++;
	if (state_lock == 0) return;
	if (state_magnet == 0) return;
	if (state_solar > 10) return;
	static struct main_timer _mt;
	if (next_move.timestamp + MOVEDEST_TIMEOUT >= stroj_now)
	{
		stroj_offset = next_move.dest - main_now; // presuneme se
		next_move.dest = stroj_now; // a vynulujeme next_move

	}
	planned_tasks_update();
	play_sound("ogg/time_transport.ogg");

//	printf("move_in_time\n");
	set_orange_led(1);
	set_lamp(1);
	
	set_green_led(1);
	set_big_red_led(1);

	_mt.handler = &_moved;
	timer_add(&_mt, main_now + 500);
	lcd_display("TRANSPORT\n");
	lcd_display("IS BEGINNING\n");

}


void planned_tasks_update(void)
{
	int i;
	for (i=0;i<_plannedtasks_n;i++)
	{
		if (_plannedtasks[i].time < stroj_now) continue;	// tenhle prikaz spustime jen pokud se presuneme do minulosti
		if ((_plannedtasks_current != _plannedtasks_n) &&	// pokud predtim nebyl zadny prikaz ke spusteni, tak zustaneme
				(_plannedtasks[i].time > _plannedtasks[_plannedtasks_current].time)) continue;	// mame-li blizsi, nezajima nas tento
		_plannedtasks_current = i;
	}
	
	if (_plannedtasks_current == _plannedtasks_n)
	{
		if (_planned_tasks_timer != NULL)
		{
			timer_del(_planned_tasks_timer);
		}
	}
	else
	{
		if (_planned_tasks_timer == NULL)
		{
			_planned_tasks_timer = xmalloc_zero(sizeof(struct main_timer));
			_planned_tasks_timer->handler = &planned_tasks_check;
		}
		timer_add(_planned_tasks_timer, STROJ_TO_REAL_TIME(_plannedtasks[_plannedtasks_current].time));
	}
}

int get_planned_tasks_count(void) {
	return _plannedtasks_n;
}
plannedtask_t * get_planned_task(int no) {
	if (no < _plannedtasks_n) return &_plannedtasks[no];
	else return NULL;
}
int planned_tasks_add(s64 time, char * command) {
	printf("adding a planned task %s at %Ld.\n",command, time);
	if (_plannedtasks_n >= 256) return -1;
	_plannedtasks[_plannedtasks_n].command = xmalloc(sizeof(char)*(strlen(command) + 2));
	strcpy(_plannedtasks[_plannedtasks_n].command, command);
	_plannedtasks[_plannedtasks_n].time = time;
	planned_tasks_update();
	return _plannedtasks_n++;	// TODO: check if this really increments the _plannedtasks_n
}
// returns the ID of that planned task -- it CAN change one or more times
// during optimizing the table
void planned_tasks_replace(int no, s64 time, char * command) {
	if (no < _plannedtasks_n)
	{
		strcpy(_plannedtasks[no].command, command);
		_plannedtasks[no].time = time;
		planned_tasks_update();
	}
}
void planned_tasks_del(int no) {
	if (no < _plannedtasks_n)
	{
		xfree(_plannedtasks[no].command);
		_plannedtasks_n--;
		_plannedtasks[no] = _plannedtasks[_plannedtasks_n];
		planned_tasks_update();
	}
}

void planned_tasks_check(struct main_timer * timer)
{
	char * command;
	int i;


	if (((get_planned_task(_plannedtasks_current)) -> time > stroj_now - PLANNED_TASKS_EXEC_THRESHOLD) &&
			((get_planned_task(_plannedtasks_current)) -> time < stroj_now + PLANNED_TASKS_EXEC_THRESHOLD))
	{
		command = xmalloc(sizeof(char) * (strlen(_plannedtasks[_plannedtasks_current].command) + 2));
		strcpy(command, _plannedtasks[_plannedtasks_current].command);
		planned_tasks_del(_plannedtasks_current);
		timer_del(timer);
		planned_tasks_update();
		loc_exec_command(command);
		xfree(command);
	}
}

char * _fortuneargs[256];

void _fortune_init(void)
{
	char * str;
	str = xmalloc(sizeof(char)*( strlen(FORTUNE_BIN) + 2));
	strcpy(str, FORTUNE_BIN);
	_fortuneargs[0] = str;
	_fortunefnames = _fortuneargs + 1;
}

void fortune_add(char * buf)
{
	int i;
	if (_fortunefnames_n >= 255) printf("Fortune file buffer full. Try to join or remove some files.\n");
	while((buf[0] == 9) || (buf[0] == 10) || (buf[0] == 13) || (buf[0] == 32))
		buf++;
	if (buf[0] == EOF) return;
	i=0;
	while((buf[i] != 9) && (buf[i] != 10) && (buf[i] != 13) && (buf[i] != 32) && (buf[i] != 0))
		i++;
	buf[i] = 0;
	char * fname;
	fname = xmalloc(sizeof(char) * (strlen(buf) + 2));
	strcpy(fname, buf);
	_fortunefnames[_fortunefnames_n] = fname;
	_fortunefnames_n++;
	printf("fortune file '%s' succesfully added.\n", fname);
}

void fortune_del(char * buf)
{
	while((buf[0] == 9) || (buf[0] == 10) || (buf[0] == 13) || (buf[0] == 32))
		buf++;
	if (buf[0] == EOF) return;
	int id = atoi(buf);
	char * old;
	old = _fortunefnames[id];
	_fortunefnames_n--;
	_fortunefnames[id] = _fortunefnames[_fortunefnames_n];
	printf("fortune file '%s' succesfully removed.\n", old);
	xfree(old);
}

void fortune_list(UNUSED char * buf)
{
	int i;
	for (i=0;i<_fortunefnames_n;i++)
		printf("%d:\t%s\n", i, _fortunefnames[i]);
}

void fortune(UNUSED char * trash) {
	int fd;
	pid_t master;
	char buf[LONG_STRING];

	int real_length;
//	printf("fortune\n");
	// pustime si ve virtualnim terminalu Fortune
	master = forkpty(&fd, NULL, NULL, NULL);
	if (!master)
	{
		file_close_all();
		execvp(FORTUNE_BIN, _fortuneargs);
	}
//	printf("master\n");
	// precteme si, co nam napsala
	real_length = read(fd, buf, LONG_STRING);
	close(fd);
	buf[real_length] = 0;
	tcp_output(buf);
	lcd_display(buf);
//	lcd_display("\n");
//	lcd_display(":o)\n");
}
void weather(s64 when) {printf("weather\n");} // given an absolute timestamp it prints the weather everywhere it can; for a weather forecast right now simply call weather(stroj_now);

void stop(UNUSED char * trash)
{
	main_shutdown = 1;
}

void mood_console(char * input)
{
	if (atoi(input) != 0)
		mood = atoi(input);
	printf("Mood: %d\n", mood);
	updatemood();
}


/* OLD FUNCTIONS -- REVISION NEEDED
// Pokud jsme zapsali do zavrene roury, musime to vedet
void sigpipe_setflag(int signal)
{
	pipe_closed = 1;
}

int _tcp_output_write(FILE * fp, char * message)
{
	pipe_closed = 0;
	fprintf(fp,"%s",message);
	if (pipe_closed)
	// Zapsali jsme do zavrene roury
	{
		// zavreme
		fclose(fp);
		return -1;
	}
	return 0;
}
void tcp_output(char * message)
{
	* Zde postihujeme vypis do telnetich klientu -- zavolanim teto funkce
	 * se prislusna zprava rozesle vsem backendum, ke kterym je pripojen
	 * klient.
	 *
	int i;
	// Zapiseme do kazdeho klienta
	for (i=0;i<output_clients_n;i++)
	{
		if (_tcp_output_write(tcp_output_fp[i],message) == -1)
		{
			// pokud se zapis nepovedl, uklidime zavrenou rouru
			tcp_output_fp[i] = tcp_output_fp[output_clients_n - 1];
			output_clients_n--;
		}
	}

	// A co treba nove roury?
	char newpipes_fn[SHORT_STRING];
	// Nacteme jejich seznam ze souboru
	FILE * newpipes_fp;
	FILE * fp_tmp;
	sprintf(newpipes_fn,"stroj-new_output-%d",master_pid);
	newpipes_fp = fopen(newpipes_fn,"r");
	while(fgets(newpipes_fn,SHORT_STRING,newpipes_fp))
	{
		// Newline na konci musime smazat, jelikoz fgets ji tam
		// nechava, ale nazev souboru je bez nej
		newpipes_fn[strlen(newpipes_fn)-1] = 0;
		// Soubor otevreme
		fp_tmp = fopen(newpipes_fn,"w");
		// Pokud se do nej jeste da psat, tak si ho ulozime i na
		// priste
		if (_tcp_output_write(fp_tmp,message) != -1)
			tcp_output_fp[output_clients_n++] = fp_tmp;

	}
	// Tim jsme vyridili vsechny mozne roury -- muzeme koncit.
	return;
}

void execute_command(void)
{
	P_TRIE cur_char;
	P_TRIE old_char;
	P_TRIE del_trie = NULL;
	* Tato funkce vezme prvni prikaz ze seznamu reprezentovaneho trii a
	 * pokusi se z nej vytriskat, co ma vlastne udelat
	 *

	char command[LONG_STRING];
	int i,j = 0;
	
	if (commands_in_trie == 0)
	{	// Pokud mame prazdny seznam, musime signalizovat chybu
		raise_error(ERR_EMPTY_LIST);
		return;
	}

	// Vytahneme z trie spravny prvek
	cur_char = commands_todo;
	old_char = NULL;
	// Pokud na tomto miste neco konci, tak je to lexikograficky mensi nez
	// cokoli, co pokracuje
	while (cur_char->ending_here == 0)
	{
		// Najdu prvni neprazdny podstrom
		for (i=0;i<256;i++)
			if (cur_char->next[i] != NULL) break;

		// ulozim si prislusny znak
		command[j] = (char) i;
		// prijde to smazat, takze mene potomku
		cur_char->descendants--;
		// a jejda, uz jiny nezbyl! No tak to musime smazat cely tento
		// strom ... (pokud uz nejsme uvnitr neho) UKLID!!
		if (cur_char->descendants == 0)
		{
			// Zrusime odkaz na sebe sama
			*(cur_char->back_ref) = 0;
			// A ulozime se ke smazani
			del_trie = cur_char;
		}
		cur_char = cur_char->next[i];
		j++;
		if (del_trie != NULL)	// Smazat??
			free(del_trie);	// Smazat!!
	}
	// a doplnime nulou
	command[j] = 0;

	// UNFINISHED

}
*/


void _add_to_trie(trie_t ** trie, char * string, void * data)
{
	if ((*trie) == NULL)
	{
		(*trie) = xmalloc_zero(sizeof(trie_t));
	}
	if (string[0] == 0)
	{
		(*trie)->data = data;
		return;
	}
	_add_to_trie(&((*trie)->child[string[0]-97]),string + 1, data);
}
