#include "stroj_casu.h"

struct info {
	int allowed_after;
	int me;
};

void parse_czech(char * command) {}
void parse_latin(char * command) {}
void parse_oldirish(char * command) {}
void parse_oldczech(char * command) {}
void parse_english(char * command) {}

trie_t *barbar_trie;

#define _B(x)			(1 << (x))
#define	BARBAR_NOTHING		_B(0)
#define BARBAR_1		_B(1)
#define BARBAR_2		_B(2)
#define BARBAR_3		_B(3)
#define BARBAR_4		_B(4)
#define	BARBAR_5		_B(5)
#define	BARBAR_MANY		_B(6)
#define	BARBAR_MORE		_B(7)
#define	BARBAR_TOOMUCH		_B(8)
#define	BARBAR_HOUR		_B(9)
#define	BARBAR_DAY		_B(10)
#define	BARBAR_YEAR		_B(11)
#define	BARBAR_HUNDRED_YEARS	_B(12)
#define	BARBAR_PLUS		_B(13)
#define	BARBAR_MINUS_1		_B(14)
#define BARBAR_NEG_VAL		_B(15)
#define	BARBAR_MOVE_IN_TIME	_B(16)
#define	BARBAR_SET_DEST_TIME	_B(17)
#define	BARBAR_FORTUNE		_B(18)
#define	BARBAR_ECHO		_B(19)
#define	BARBAR_WAIT		_B(20)
#define	BARBAR_EXEC		_B(21)
#define BARBAR_PLEASE		_B(22)
#define BARBAR_ALL		0x007fffff
#define BARBAR_NUMS		(BARBAR_1 | BARBAR_2 | BARBAR_3 | BARBAR_4 | BARBAR_5 | BARBAR_MANY)
#define BARBAR_TIME		(BARBAR_HOUR | BARBAR_DAY | BARBAR_YEAR | BARBAR_HUNDRED_YEARS)
#define BARBAR_ALG_OPERATORS	(BARBAR_PLUS | BARBAR_NEG_VAL | BARBAR_MINUS_1) 
#define BARBAR_FUNCS		(BARBAR_MOVE_IN_TIME | BARBAR_SET_DEST_TIME | BARBAR_FORTUNE | BARBAR_ECHO | BARBAR_WAIT)

void _parse_barbar_init(void)
{
	// This function has to be called ONLY ONCE and from the global init function
	// It creates the environment for the barbar language parser
	//
	

	const char * known_strings[] = {
		"p", // do nothing
		"i", "ii", "iii", "iiii", "iiiii", "uf", "uuf", "uuuf", // 1-5, many, more, too much
		"t", "tt", "ttt", "tttt", // hour, day, year, hundred years
		"y", "ee", "ou", //  plus, subtract 1, negative value
		"vuuum", // Move in time
		"um", // Set the time, where to move
		"juuu", // Fortune
		"ui", // Echo
		"zzzz", // wait for a specified time
		"hm", // do command and keep on reading the next command
		"ya", // Please
	};
#define ___BARBAR_KNOWN_STRINGS_NO	23	////// PRASO -- nevite nekdo, jak sehnat pocet prvku v poli?

	struct info data[] = {
		{BARBAR_ALL, BARBAR_NOTHING},
		{(BARBAR_NUMS | BARBAR_ALG_OPERATORS | BARBAR_TIME | BARBAR_EXEC), BARBAR_1},
		{(BARBAR_NUMS | BARBAR_ALG_OPERATORS | BARBAR_TIME | BARBAR_EXEC), BARBAR_2},
		{(BARBAR_NUMS | BARBAR_ALG_OPERATORS | BARBAR_TIME | BARBAR_EXEC), BARBAR_3},
		{(BARBAR_NUMS | BARBAR_ALG_OPERATORS | BARBAR_TIME | BARBAR_EXEC), BARBAR_4},
		{(BARBAR_NUMS | BARBAR_ALG_OPERATORS | BARBAR_TIME | BARBAR_EXEC), BARBAR_5},
		{(BARBAR_NUMS | BARBAR_ALG_OPERATORS | BARBAR_TIME | BARBAR_EXEC), BARBAR_MANY},
		{(BARBAR_NUMS | BARBAR_ALG_OPERATORS | BARBAR_TIME | BARBAR_EXEC), BARBAR_MORE},
		{(BARBAR_NUMS | BARBAR_ALG_OPERATORS | BARBAR_TIME | BARBAR_EXEC), BARBAR_TOOMUCH},
		{(BARBAR_NUMS | BARBAR_EXEC),BARBAR_HOUR},
		{(BARBAR_NUMS | BARBAR_EXEC),BARBAR_DAY},
		{(BARBAR_NUMS | BARBAR_EXEC),BARBAR_YEAR},
		{(BARBAR_NUMS | BARBAR_EXEC),BARBAR_HUNDRED_YEARS},
		{(BARBAR_NUMS | BARBAR_EXEC),BARBAR_PLUS},
		{(BARBAR_NUMS | BARBAR_EXEC | BARBAR_TIME | BARBAR_ALG_OPERATORS),BARBAR_MINUS_1},
		{(BARBAR_NUMS | BARBAR_EXEC),BARBAR_NEG_VAL},
		{(BARBAR_EXEC),BARBAR_MOVE_IN_TIME},
		{(BARBAR_NUMS),BARBAR_SET_DEST_TIME},
		{(BARBAR_EXEC),BARBAR_FORTUNE},
		{(BARBAR_EXEC),BARBAR_ECHO},
		{(BARBAR_NUMS),BARBAR_WAIT},
		{(BARBAR_ALL),BARBAR_EXEC},
		{(BARBAR_ALL),BARBAR_PLEASE}
	};
	
	int i;

	struct info * data_transport;

	for (i=0;i<___BARBAR_KNOWN_STRINGS_NO;i++)
	{
		data_transport = xmalloc(sizeof(struct info));
		*data_transport = data[i];
		_add_to_trie(&barbar_trie, known_strings[i], data_transport);
	}

}

void _barbar_error(void)
{
	tcp_output("Grrr!\n");
	lcd_display("Grrr!\n");
	play_sound("ogg/grrr.ogg");
}

#define BARBAR_INF_NUM		1
#define BARBAR_INF_HOURS	2
#define BARBAR_INF_DAYS		4
#define BARBAR_INF_YEARS	8

typedef struct _barbar_num_time {
	int num;
	int hours;
	int days;
	int years;
	int inf_flag;
} _barbar_num_time_t;

#define _barbar_num_time_timestamp(x)	((((x).years*36524 + (x).days*100)*24 + (x).hours*100)*3600*10 + stroj_now)
#define _barbar_const_to_num(x)		(((x) == BARBAR_1) ? 1 : (((x) == BARBAR_2) ? 2 : (((x) == BARBAR_3) ? 3 : (((x) == BARBAR_4) ? 4 : (((x) == BARBAR_5) ? 5 : -1)))))

int _barbar_wait_flag;

_barbar_num_time_t _barbar_num_time_process(int * command, int n)
{
	_barbar_num_time_t num_time = {0,0,0,0,0};

	int num_old = 0;
	int num_now = 0;
	int inf_flag = 0;
	int i;

	if (n == 0) return num_time;
	if (n == 1)
	{
		num_time.num = _barbar_const_to_num(command[0]);
		if (num_time.num == -1) num_time.num = 6; // if "many", put >5 into num -- it gets translated into many
		return num_time;
	}
	for (i=0;i<n;i++)
	{
		if (command[i] == BARBAR_HOUR)
		{
			num_time.hours += num_old;
			num_old = 0;
			if (inf_flag)
				num_time.inf_flag |= BARBAR_INF_DAYS;
			inf_flag = 0;
			continue;
		}
		if (command[i] == BARBAR_DAY)
		{
			num_time.days += num_old;
			num_old = 0;
			if (inf_flag)
				num_time.inf_flag |= BARBAR_INF_HOURS;
			inf_flag = 0;
			continue;
		}
		if (command[i] == BARBAR_YEAR)
		{
			num_time.years += num_old;
			num_old = 0;
			if (inf_flag)
				num_time.inf_flag |= BARBAR_INF_YEARS;
			inf_flag = 0;
			continue;
		}
		if (command[i] == BARBAR_HUNDRED_YEARS)
		{
			num_time.years += num_old * 100;
			num_old = 0;
			if (inf_flag)
				num_time.inf_flag |= BARBAR_INF_YEARS;
			inf_flag = 0;
			continue;
		}
		if (command[i] == BARBAR_PLUS) continue;
		if (command[i] == BARBAR_MINUS_1)
		{
			num_old--;
			continue;
		}
		if (command[i] == BARBAR_NEG_VAL) inf_flag = -(1 + inf_flag);

		num_now = _barbar_const_to_num(command[i]);
		if (num_now == -1)
		{
			inf_flag = 1;
			num_now = 6;
		}
		if (inf_flag < 0)
		{
			num_now = -num_now;
			inf_flag = -(1 + inf_flag);
		}
		num_old += num_now;
	}
	
	if (inf_flag)
		num_time.inf_flag |= BARBAR_INF_NUM;

	num_time.num = num_old;
	return num_time;
}

void _barbar_exec(int * command, int n)
{
	int mood_ok;
	int i;
	_barbar_num_time_t num_time;

	char outstr[LONG_STRING] = "";


	if((n > 0) && (command[0] == BARBAR_PLEASE))
	{
		mood_ok = please();
		command++;
		n--;
	}
	else
		mood_ok = noplease();

	if (!mood_ok)
	{
		_barbar_error();
		return;
	}
	if (n == 0) return;

	if (command[0] == BARBAR_MOVE_IN_TIME)
	{
		move_in_time();
		return;
	}
	if (command[0] == BARBAR_FORTUNE)
	{
		fortune(NULL);
		return;
	}
	if (command[0] == BARBAR_ECHO)
	{
		tcp_output("");
		lcd_display("");
		return;
	}
	if (command[0] == BARBAR_WAIT)
	{
		num_time = _barbar_num_time_process(command + 1, n-1);
		if (num_time.num == 0)
		{
			// Zbytek prikazu se odlozi za vlozeny casovy usek
			_barbar_wait_flag = _barbar_num_time_timestamp(num_time);
			printf("%Ld", _barbar_wait_flag);
		}
		else
			_barbar_error();
		return;

	}
	if (command[0] == BARBAR_SET_DEST_TIME)
	{
		num_time = _barbar_num_time_process(command + 1,n-1);
		if (num_time.num == 0)
			set_next_move(_barbar_num_time_timestamp(num_time));
		else
			_barbar_error();
		return;
	}
	num_time = _barbar_num_time_process(command, n);
	i=num_time.hours + (num_time.inf_flag & BARBAR_INF_HOURS);

	if ((num_time.inf_flag & BARBAR_INF_HOURS) || num_time.hours > 5)
		strcat(outstr, "uf");
	else
	{
		while (num_time.hours > 0)
		{
			strcat(outstr, "I");
			num_time.hours--;
		}
		strcat(outstr, " ");
	}
	if (i) strcat(outstr, "t ");
	i=num_time.days + (num_time.inf_flag & BARBAR_INF_DAYS);

	if ((num_time.inf_flag & BARBAR_INF_DAYS) || num_time.days > 5)
		strcat(outstr, "uf");
	else
	{
		while (num_time.days > 0)
		{
			strcat(outstr, "I");
			num_time.days--;
		}
		strcat(outstr, " ");
	}
	if (i) strcat(outstr, "tt ");
	i=num_time.years + (num_time.inf_flag & BARBAR_INF_YEARS);

	if ((num_time.inf_flag & BARBAR_INF_YEARS) || num_time.years > 5)
		strcat(outstr, "uf");
	else
	{
		while (num_time.years > 0)
		{
			strcat(outstr, "I");
			num_time.years--;
		}
		strcat(outstr, " ");
	}
	if (i) strcat(outstr, "ttt ");

	if ((num_time.inf_flag & BARBAR_INF_NUM) || num_time.num > 5)
		strcat(outstr, "uf");
	else
	{
		while (num_time.num > 0)
		{
			strcat(outstr, "I");
			num_time.num--;
		}
		strcat(outstr, " ");
	}
	strcat(outstr, "\n");
	tcp_output(outstr);
	lcd_display(outstr);
//	printf("%d %d %d %d %d\n",num_time.num, num_time.hours, num_time.days, num_time.years, num_time.inf_flag);
}

void parse_barbar(char * command)
{
	char * command_local;
	int * _command;
	int i = 0;
	int j = 0;
	int mood_ok;
	trie_t * now_in;
	now_in = barbar_trie;
	int parsed_command[LONG_STRING];
	int allowed_toks = BARBAR_ALL;

	_barbar_wait_flag = 0;
	// Ulozime si mistni verzi
	command_local = xmalloc_zero(strlen(command)*sizeof(char));
	strcpy(command_local, command);
	// prevedeme na mala pismena
	for (i=0;i<strlen(command_local);i++)
		command_local[i] = (char) tolower((int) command_local[i]);

	i=0;
	// a cyklime po pismenkach
	while (command_local[i] != 0)
	{
		if (_barbar_wait_flag)
		{
			planned_tasks_add(_barbar_wait_flag, command_local + i);
			return;
		}
		if (parsed_command[j-1] == BARBAR_EXEC) // pokus o provedeni prikazu
		{
			_command = xmalloc(j*sizeof(int));
			memcpy(_command, parsed_command, j*sizeof(int));
			_barbar_exec(_command, j);
			j=0;
			allowed_toks = BARBAR_ALL;
			xfree(_command);
		}
		if (parsed_command[j-1] == BARBAR_ECHO) // Echo vypise zbytek
		{
			if (j >= 2 && parsed_command[j-2] == BARBAR_PLEASE)
				mood_ok = please();
			else
				mood_ok = noplease();

			if (mood_ok)
			{
				tcp_output(command_local + i);
				lcd_display(command_local + i);
			}
			else
				_barbar_error();
			return;
		}
		switch(command_local[i])
		{
			// bily znak
			case 32:
			case 9:
			case 10:
			case 13:
				if (now_in == barbar_trie)
				{// na zacatku nebo po jinem bilem znaku ho polkneme
					i++;
					continue;
				}
				if (now_in->data == NULL)
				{// konec slova, ale toto slovo nezname -> chyba
					xfree(command_local);
					_barbar_error();
					return;
				}
				if ((allowed_toks & ((struct info *) now_in->data)->me) == 0)
				{// konec slova, ktere sice zname, ale nesmi byt na tomto miste -> chyba
					xfree(command_local);
					_barbar_error();
					return;
				}
				// konec slova, ktere tu smi byt
				parsed_command[j++] = ((struct info *) now_in->data)->me;
				allowed_toks = ((struct info *) now_in->data)->allowed_after;
				now_in = barbar_trie;
				break;
			// jine pismeno
			default:
				// neni pismeno
				if (command_local[i] - 97 < 0)
					break;
				if (command_local[i] - 97 >= 26)
					break;
				// neplatne slovo
				if (now_in->child[command_local[i] - 97] == NULL)
				{
					xfree(command_local);
					_barbar_error();
					return;
				}
				// zatim stale platne slovo
				now_in = now_in->child[command_local[i] - 97];


		}
		i++;
	}
	// pokus o provedeni prikazu
	_command = xmalloc(j*sizeof(int));
	memcpy(_command, parsed_command, j*sizeof(int));
	_barbar_exec(_command, j);
	xfree(_command);
	xfree(command_local);
}

